#!/usr/bin/env node

import * as THREE from "three"
import * as OBC from "@thatopen/components"
import * as fs from "fs"
import * as path from "path"
import { select } from '@inquirer/prompts';

// To have something as reusable as possible, define a function
// that returns all files that matches an extension in the
// folder where the code is executed.
function getFiles(extension) {
	// process.cwd means the current working directory where the code is executed.
	// During development, the folder will be the root directory.
  return fs.readdirSync(process.cwd()).filter(file => path.extname(file).toLowerCase() === extension.toLowerCase())
}

// The code relies in one IFC and one IDS file to work.
// Is the user choice which files to use, so we need to make a prompt using inquirer
const promptUser = async () => {
  const ifcFiles = getFiles('.ifc');
  const idsFiles = getFiles('.ids');

  if (ifcFiles.length === 0 || idsFiles.length === 0) {
    console.log("No IFC or IDS files found in the current directory.");
    return;
  }

  const ifcFile = await select({ message: "Select the IFC file:", choices: ifcFiles })
  const idsFile = await select({ message: "Select the IDS file:", choices: idsFiles })

	// As getting the content from a file involves to have the full path
	// we join the current working directory with the name of the file chosen.
  const ifcPath = path.join(process.cwd(), ifcFile)
  const idsPath = path.join(process.cwd(), idsFile)

  return { ifcPath, idsPath }
}

// This is the main function to check the requirements in the IDS file
// against the IFC model
const checkIFC = async () => {
  const selectedFiles = await promptUser();
  if (!selectedFiles) return

	// We read the IFC file as binary data as is easier to convert it to
	// a Uint8Array needed by the IfcLoader.
	// However, it can also be read as text.
  const ifc = fs.readFileSync(selectedFiles.ifcPath)
  
  // The IDS file is always read as text, thus the utf8
  const ids = fs.readFileSync(selectedFiles.idsPath, "utf8")

  const components = new OBC.Components()

  const ifcLoader = components.get(OBC.IfcLoader)
  const ifcBuffer = new Uint8Array(ifc.buffer)
  const model = await ifcLoader.load(ifcBuffer)

  // It's important to process the model relations, so getting
  // psets, materials, classifications, etc, required by
  // some facets in IDS files is possible.
  const indexer = components.get(OBC.IfcRelationsIndexer)
  await indexer.process(model)

	// Using the brand new IDSSpecifications component, we load the ids.
  const idsSpecs = components.get(OBC.IDSSpecifications)
  
  // One single IDS may have multiple specifications.
  const specs = idsSpecs.load(ids)

	// As the goal is to create a BCF file, we initialize
	// the BCFTopics component.
  const topics = components.get(OBC.BCFTopics)
  topics.setup()

	// As the topic will have a viewpoint with the failing
	// elements, then a world is needed.
  const world = components.get(OBC.Worlds).create()
  const viewpoints = components.get(OBC.Viewpoints)

  for (const [index, spec] of specs.entries()) {
    const result = await spec.test(model)
    
    // From the specification test result, we get the list of guid
    // from the elements that fail.
    const failingGuids = result
      .filter(check => !check.pass)
      .map(check => check.guid)
      .filter(guid => guid)

    const topic = topics.create({
      title: "Failing Information Requirements",
      description: spec.description
    })

    const viewpoint = new OBC.Viewpoint(components, world, { setCamera: false })
    viewpoints.list.set(viewpoint.guid, viewpoint)

		// In the viewpoint, we need to add the failing guids to the selection
		// components. This makes possible for the target software of the BCF
		// to select the elements
    viewpoint.selectionComponents.add(...failingGuids)
    
    // Also, as another input for the user using the BCF, we colorize the 
    // failing elements with red.
    viewpoint.componentColors.set(new THREE.Color("red"), failingGuids)

    topic.viewpoints.add(viewpoint.guid)

    const bcfBlob = await topics.export([topic])
    
    // topics.export returns a blob, which is imperative to the web.
    // We can't use blob in a NodeJS environment, so we take the arrayBuffer
    // from it and convert it to a Buffer.
    const bcfData = Buffer.from(await bcfBlob.arrayBuffer())
    const exportPath = path.join(process.cwd(), `result-${index + 1}.bcf`)
    fs.writeFileSync(exportPath, bcfData)
  }
}

// Do not forget to execute the main function
checkIFC()